#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Low level classes and methods for io
"""
from collections import OrderedDict as od
from os.path import join, exists, basename, splitext
from pyaerocom import const
from pyaerocom.exceptions import FileConventionError
try:
    from ConfigParser import ConfigParser
except: 
    from configparser import ConfigParser

class FileConventionRead(object):
    """Class that represents a file naming convention for reading Aerocom files
    
    Attributes
    ----------
    name : str
        name of this convention (e.g. "aerocom3")
    file_sep : str
        filename delimiter for accessing different variables
    year_pos : int
        position of year information in filename after splitting using 
        delimiter :attr:`file_sep` 
    var_pos : int
        position of variable information in filename after splitting using 
        delimiter :attr:`file_sep` 
    ts_pos : int
        position of information of temporal resolution in filename after 
        splitting using delimiter :attr:`file_sep` 
    vert_pos : int
        position of information about vertical resolution of data
    data_id_pos : int
        position of data ID 
    """
    _io_opts = const
    AEROCOM3_VERT_INFO = {'2d'  : ['surface', 'column', 'modellevel'],
                          '3d'  : ['modellevelatstations']}
    
    def __init__(self, name="aerocom3", file_sep="_", year_pos=None,
                 var_pos=None, ts_pos=None, vert_pos=None, data_id_pos=None,
                 from_file=None):
       
        self.name = name
        self.file_sep = file_sep

        self.year_pos = year_pos
        self.var_pos = var_pos
        self.ts_pos = ts_pos
        self.vert_pos = vert_pos
        self.data_id_pos = data_id_pos
        
        if from_file is not None:
            self.from_file(from_file)
        else:
            try:
                self.import_default(self.name) 
            except:
                pass
      
    @property
    def info_init(self):
        """Empty dictionary containing init values of infos to be 
        extracted from filenames
        """
        return od(year=None, var_name=None, ts_type=None, vert_code='',
                  is_at_stations=False, data_id='')
        
    def from_file(self, file):
        """Identify convention from a file
        
        Currently only two conventions (aerocom2 and aerocom3) exist that are
        identified by the delimiter used.
        
        Parameters
        ----------
        file : str
            file path or file name
        
        Returns
        -------
        FileConventionRead
            this object (with updated convention)
            
        Raises
        ------
        FileConventionError
            if convention cannot be identified
            
        Example
        -------
        >>> from pyaerocom.io import FileConventionRead
        >>> filename = 'aerocom3_CAM5.3-Oslo_AP3-CTRL2016-PD_od550aer_Column_2010_monthly.nc'
        >>> print(FileConventionRead().from_file(filename))
        pyaeorocom FileConventionRead
        name: aerocom3
        file_sep: _
        year_pos: -2
        var_pos: -4
        ts_pos: -1
        """
        
        if basename(file).count("_") >= 4:
            self.import_default("aerocom3")
        elif basename(file).count(".") >= 4:
            self.import_default("aerocom2")
        else:
            raise FileConventionError('Could not identify convention from '
                                      'input file {}'.format(basename(file)))
        self.check_validity(file)
        return self
    
    def check_validity(self, file):
        """Check if filename is valid"""
        info = self.get_info_from_file(file)
        year = info["year"]
        if not info['ts_type'] in self._io_opts.GRID_IO.TS_TYPES:
            raise FileConventionError("Invalid ts_type %s in filename %s"
                                           %(info['ts_type'], basename(file)))
        elif not (const.MIN_YEAR <= year <= const.MAX_YEAR):
            raise FileConventionError("Invalid year %d in filename %s"
                          %(info['year'], basename(file)))
    
    def _info_from_aerocom3(self, file):
        """Extract info from filename Aerocom 3 convention
        
        Parameters
        -----------
        file : str
            netcdf file name
            
        Returns
        -------
        dict
            dictionary containing infos that were extracted from filename
        """
        # init result dictionary
        info = self.info_init
        
        spl = splitext(basename(file))[0].split(self.file_sep)
        # phase 3 file naming convention
        try:
            info["year"] = int(spl[self.year_pos])
        except:
            msg = ("Failed to extract year information from file {} "
                   "using file convention Aerocom 3".format(basename(file), 
                                                            self.name))
            raise FileConventionError(msg)
        try:
            # include vars for the surface
            if spl[self.vert_pos].lower() in self.AEROCOM3_VERT_INFO['2d']:
                info["var_name"] = spl[self.var_pos]
            # also include 3d vars that provide station based data
            # and contain the string vmr in this case the variable name has to 
            # be slightly changed to the  aerocom phase 2 naming
            elif spl[self.vert_pos].lower() in self.AEROCOM3_VERT_INFO['3d']:
                if 'vmr' in spl[self.var_pos]:
                    info['var_name'] = spl[self.var_pos].replace('vmr', 'vmr3d')  
                else:
                    info['var_name'] = spl[self.var_pos]
            else:
                raise FileConventionError('Invalid file name (Aerocom 3 '
                                          'naming convention).\n'
                                          '{}\nInvalid string identifier for '
                                          'vertical coordinate: {}'
                                          .format(file, spl[self.vert_pos]))
        except Exception as e:
            raise FileConventionError('Failed to extract variable name from '
                                      'file {} using file convention {}.\n'
                                      'Error: {}'.format(basename(file), 
                                                         self.name, repr(e)))
        try:
            info["ts_type"] = spl[self.ts_pos]
        except:
            raise FileConventionError('Failed to extract ts_type from '
                                      'file {} using file convention {}' 
                                      .format(basename(file), self.name))
        try:
            info["vert_code"] = spl[self.vert_pos]
        except:
            raise FileConventionError('Failed to extract vert_code from '
                                      'file {} using file convention {}' 
                                      .format(basename(file), self.name))
            
        try:
            info["data_id"] = self.file_sep.join(spl[self.data_id_pos:self.var_pos])
        except:
            raise FileConventionError('Failed to extract model name from '
                                      'file {} using file convention {}' 
                                      .format(basename(file), self.name))
        if'atstations' in file.lower():
            info['is_at_stations'] = True
        return info
    
    def _info_from_aerocom2(self, file):
        """Extract info from filename Aerocom 2 convention
        
        Parameters
        -----------
        file : str
            netcdf file name
            
        Returns
        -------
        dict
            dictionary containing infos that were extracted from filename
        """
        info = self.info_init
        if self.file_sep == ".":
            spl = basename(file).split(self.file_sep)
        else:
            spl = splitext(basename(file))[0].split(self.file_sep)
        try:
            info["year"] = int(spl[self.year_pos])
        except:
            raise FileConventionError('Failed to extract year information '
                                      'from file {} using file '
                                      'convention {}' 
                                      .format(basename(file), self.name))
        try:
            info["var_name"] = spl[self.var_pos]
        except:
            raise FileConventionError('Failed to extract variable information '
                                      'from file {} using file '
                                      'convention {}' 
                                      .format(basename(file), self.name))
        try:
            info["ts_type"] = spl[self.ts_pos]
        except:
            raise FileConventionError('Failed to extract ts_type '
                                      'from file {} using file '
                                      'convention {}' 
                                      .format(basename(file), self.name))

        try:
            info['data_id'] = '.'.join(spl[self.data_id_pos:self.ts_pos])
        except:
            raise FileConventionError('Failed to extract name '
                                      'from file {} using file '
                                      'convention {}' 
                                      .format(basename(file), self.name))
        if 'atstations' in file.lower():
            raise Exception('Developers: please debug (file convention '
                            'Aerocom 2 should not have atstations '
                            'encoded in file name)')
        return info
    
    def get_info_from_file(self, file):
        """Identify convention from a file
        
        Currently only two conventions (aerocom2 and aerocom3) exist that are
        identified by the delimiter used.
        
        Parameters
        ----------
        file : str
            file path or file name
        
        Returns
        -------
        OrderedDict
            dictionary containing keys `year, var_name, ts_type` and 
            corresponding variables, extracted from the filename 
            
        Raises
        ------
        FileConventionError
            if convention cannot be identified
            
        Example
        -------
        >>> from pyaerocom.io import FileConventionRead
        >>> filename = 'aerocom3_CAM5.3-Oslo_AP3-CTRL2016-PD_od550aer_Column_2010_monthly.nc'
        >>> conv = FileConventionRead("aerocom3")
        >>> info = conv.get_info_from_file(filename)
        >>> for item in info.items(): print(item)
        ('year', 2010)
        ('var_name', 'od550aer')
        ('ts_type', 'monthly')
        """
        if self.name == 'aerocom3':
            return self._info_from_aerocom3(file)
        elif self.name == 'aerocom2':
            return self._info_from_aerocom2(file)
        
    
    
    def string_mask(self, data_id, var, year, ts_type, vert_which=None):
        """Returns mask that can be used to identify files of this convention
        
        Parameters
        ----------
        data_id : str
            experiment ID (e.g. GISS-MATRIX.A2.CTRL)
        var : str
            variable string ID (e.g. "od550aer")
        year : int
            desired year of observation (e.g. 2012)
        ts_type : str
            string specifying temporal resolution (e.g. "daily")
        
        Example
        -------
            
        import re
        conf_aero2 = FileConventionRead(name="aerocom2")
        conf_aero3 = FileConventionRead(name="aerocom2")
        
        var = od550aer
        year = 2012
        ts_type = "daily"
        
        match_str_aero2 = conf_aero2.string_mask(var, year, ts_type)
        
        match_str_aero3 = conf_aero3.string_mask(var, year, ts_type)
            
        """
        if ts_type is None:
            ts_type = '*'
        if self.name == "aerocom2":
            if vert_which is not None:
                raise FileConventionError('Specification of vert_which ({}) is '
                                          'not supported for '
                                          'aerocom2 naming convention'
                                          .format(vert_which))
                
            return ".".join(['.*', data_id, ts_type, var, str(year), 'nc'])
        elif self.name == "aerocom3":
            if vert_which is None:
                vert_which = '.*'
            return "_".join(['.*',  data_id, var, vert_which, str(year), ts_type]) + '.nc'
        else:
            raise NotImplementedError("File matching mask for convention %s "
                                      "not yet defined..." %self.name)
            
    def import_default(self, name):
        """Checks and load default information from database"""
        from pyaerocom import __dir__
        fpath = join(__dir__, "data", "file_conventions.ini")
        if not exists(fpath):
            raise IOError("File conventions ini file could not be found: %s"
                          %fpath)
        conf_reader = ConfigParser()
        conf_reader.read(fpath)
        if not name in conf_reader:
            raise NameError("No default available for %s" %name)
        self.name = name
        for key, val in conf_reader[name].items():
            if key in self.__dict__:
                try:
                    val = int(val)
                except:
                    pass
                self.__dict__[key] = val
                
    
    def from_dict(self, new_vals):
        """Load info from dictionary
        
        Parameters
        ----------
        new_vals : dict
            dictionary containing information
        
        Returns
        -------
        self
        """
        for k, v in new_vals.items():
            if k in self.__dict__:
                self.__dict__[k] = v
        return self
    
    def to_dict(self):
        """Convert this object to ordered dictionary"""
        return od(name = self.name,
                  file_sep = self.file_sep,
                  year_pos = self.year_pos,
                  var_pos = self.var_pos,
                  ts_pos = self.ts_pos,
                  vert_pos = self.vert_pos,
                  data_id_pos = self.data_id_pos)
      
    def __repr__(self):
       return ("%s %s" %(self.name, super(FileConventionRead, self).__repr__()))
   
    def __str__(self):
        s = "\npyaeorocom FileConventionRead"
        for k, v in self.to_dict().items():
            s += "\n%s: %s" %(k, v)
        return s
    
if __name__=="__main__":
    conf = FileConventionRead()
    
    print(conf)
        
    d = od(name = "Fake",
           file_sep = 10,
           year_pos = -6,
           var_pos = 15,
           ts_pos = 3)
    print(conf.from_dict(d))
    try:
        conf.import_default("blaaa")
    except NameError:
        print("Works as expected")
    conf.import_default("aerocom3")
    print(conf)
    
    conf = FileConventionRead(name="aerocom2")
    print(conf)
    
    fname = 'aerocom3_TM5_AP3-INSITU_vmrch4_ModelLevelAtStations_2010_monthly.nc'
    
    print('\nFrom file: {}'.format(conf.from_file(fname)))
    print(conf.get_info_from_file(fname))
    
    ff = FileConventionRead(from_file='aerocom.CALIOP3.monthly.ec5323Ddust.2006.nc')
    print(ff)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Collection of station data classes
"""
import numpy as np
from collections import OrderedDict as od
from pyaerocom._lowlevel_helpers import dict_to_str, list_to_shortstr, BrowseDict

class VerticalProfile(object):
    """Object for single variable profile data
        
    Attributes
    ----------
    data : :obj:`ndarray` (or similar), optional
        profile data
    altitude : :obj:`ndarray` (or similar), optional
        altitude levels of profile
    var_name : str
        name of variable
    
    """
    def __init__(self, data=None, altitude=None, dtime=None,
                 var_name=None, data_err=None, var_unit=None, 
                 altitude_unit=None, **location_info):
        
        if data is None:
            data = []
        if data_err is None:
            data_err = []
        if dtime is None:
            dtime = []
        if altitude is None:
            altitude = []
        if var_name is None:
            var_name = 'data'
            
        self._var_name = None
        self._data = []
        self._data_err = []
        self._altitude = []
        self._vert_coord_name = None
        self._vert_coord_vals = od()
        
        self.var_info = BrowseDict()
        self.var_info['altitude'] = od()
        
        self.update(**location_info)
        
        self.var_name = var_name
        self.dtime = dtime
        self.data = data
        self.data_err = data_err
        self.altitude = altitude

        if var_unit is not None:
            self.var_unit = var_unit
        if altitude_unit is not None:
            self.altitude_unit = altitude_unit
      
    def update(self, **kwargs):
        for k, v in kwargs.items():
            self.__dict__[k] = v
    
    @property
    def var_name(self):
        """Variable name of profile data
        
        Raises
        ------
        ValueError
            if variable name is not available
        """
        if self._var_name  is None:
            raise ValueError('Variable name is not assigned...')
        return self._var_name
    
    @var_name.setter
    def var_name(self, val):
        if not isinstance(val, str):
            raise ValueError('Cannot assign {} as variable name, need string')
        if not val in self.var_info:
            self.var_info[val] = od()
        self._var_name = val
            
    @property
    def data(self):
        """Array containing data values corresponding to data"""
        return np.float64(self._data)
    
    @data.setter
    def data(self, val):
        self._data = val
    
    @property
    def data_err(self):
        """Array containing data values corresponding to data"""
        return np.float64(self._data_err)
    
    @data_err.setter
    def data_err(self, val):
        self._data_err = val
        
    @property
    def altitude(self):
        """Array containing altitude values corresponding to data"""
        if len(self._altitude) == len(self._data):
            return np.float64(self._altitude)
        return self.compute_altitude()
    
    @altitude.setter
    def altitude(self, val):
        self._altitude = val
        
    @property
    def var_unit(self):
        """Unit of variable (requires var_name to be available)"""
        if not 'units' in self.var_info[self.var_name]:
            add_str=''
            if 'unit' in self.var_info[self.var_name]:
                add_str = 'Note: attribute name "unit" is deprecated. Please ' \
                          'use "units"'
            raise ValueError('Unit is not defined. {}'.format(add_str))
        return self.var_info[self.var_name]['units']
            
    @var_unit.setter
    def var_unit(self, val):
        self.var_info[self.var_name]['units'] = val
    
    @property
    def altitude_unit(self):
        """Unit of altitude"""
        if not 'units' in self.var_info['altitude']:
            raise ValueError('Altitude unit is not defined')
        return self.var_info['altitude']['units']
            
    @altitude_unit.setter
    def altitude_unit(self, val):
        self.var_info['altitude']['units'] = val
        
    def compute_altitude(self):
        """Compute altitude based on vertical coorinate information"""
        raise NotImplementedError
        from pyaerocom.vert_coords import _VertCoordConverter as conv
        if not self.vert_coord_info:
            raise ValueError('No information about vertical coordinate found')
        elif not self.vert_coord_name in conv.supported:
            raise ValueError('Name of vertical coordinate not registered')
        
    def plot(self, plot_errs=True, whole_alt_range=False, rot_xlabels=30, 
             errs_shaded=True, errs_alpha=0.1, add_vertbar_zero=True,
             **kwargs):
        """Simple plot method for vertical profile
        
        Parameters
        ----------
        plot_errs : bool
            if True, and if errordata is available
            
        """
        import matplotlib.pyplot as plt
        if 'figsize' in kwargs:
            figsize = kwargs.pop('figsize')
        else:
            figsize = (4, 8)
        if 'ax' in kwargs:
            ax = kwargs.pop('ax')
        else:
            _, ax = plt.subplots(1,1, figsize=figsize)
        p = ax.plot(self.data, self.altitude, '-x', **kwargs)
        
        c = p[0].get_color()
        if rot_xlabels:
            for lbl in ax.get_xticklabels():
                lbl.set_rotation(rot_xlabels)
                
        xlab = self.var_name
        ylab = 'Altitude'
        try:
            xlab += ' [{}]'.format(self.var_info[self.var_name]['units'])
        except:
            pass
        
        try:
            ylab += ' [{}]'.format(self.var_info['altitude']['units'])
        except:
            pass
        ax.set_xlabel(xlab)
        ax.set_ylabel(ylab)
        
        if whole_alt_range:
            ax.set_ylim([np.min([0, self.altitude.min()]), self.altitude.max()])
        if plot_errs and len(self.data_err) == len(self.data):
                
            lower = self.data - self.data_err
            upper = self.data + self.data_err
            if errs_shaded:            
                ax.fill_betweenx(self.altitude, lower, upper, 
                                 color=c, alpha=errs_alpha)
            else:
                ax.errorbar(self.data, self.altitude, xerr=self.data_err, 
                            ls=' ', marker=' ', color='#cccccc')
        if add_vertbar_zero:
            xl = ax.get_xlim()
            if xl[0] < 0 < xl[1]:
                ax.plot([0,0], ax.get_ylim(), '--', color='#cccccc')
        ax.figure.tight_layout()
        
        return ax

    def __len__(self):
        return len(self['data'])
    
    def __str__(self):
        head = "Pyaerocom {}".format(type(self).__name__)
        s = "\n{}\n{}".format(head, len(head)*"-")
        s += ('\naltitude: {}'
              '\ndata: {}'
              '\ndata_err: {}'
              .format(list_to_shortstr(self.altitude),
                      list_to_shortstr(self.data), 
                      list_to_shortstr(self.data_err)))
        arrays = ''
        for k, v in self.__dict__.items():
            if k.startswith('_'):
                continue
            if isinstance(v, dict) and len(v) > 0:
                s += "\n{} (dict)".format(k)
                s = dict_to_str(v, s)
            elif isinstance(v, list):
                s += "\n{} (list, {} items)".format(k, len(v))
                s += list_to_shortstr(v)
            elif isinstance(v, np.ndarray) and v.ndim==1:
                arrays += "\n{} (array, {} items)".format(k, len(v))
                arrays += list_to_shortstr(v)
            else:
                s += "\n%s: %s" %(k,v)
        s += arrays
        return s
    
if __name__=="__main__":
    
    d = data=[1,2,3,4,3,2,1,1,1,1]
    p = VerticalProfile(data=d,
                        altitude=np.arange(len(d)),
                        data_err=np.ones(len(d))*.3,
                        var_unit='1/Mm',
                        altitude_unit='m')
    p.plot()
    
    
        
        
        




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module containing time resampling functionality
"""
import pandas as pd
import xarray as xarr
from pyaerocom import const
from pyaerocom.exceptions import TemporalResolutionError
from pyaerocom.tstype import TsType
from pyaerocom.time_config import TS_TYPE_TO_PANDAS_FREQ
from pyaerocom.helpers import (resample_time_dataarray, 
                               resample_timeseries, isnumeric)

class TimeResampler(object):
    """Object that can be use to resample timeseries data
    
    It supports hierarchical resampling of :class:`xarray.DataArray` objects 
    and :class:`pandas.Series` objects.
    
    Hierarchical means, that resampling constraints can be applied for each 
    level, that is, if hourly data is to be resampled to monthly, it may be 
    specified to first required minimum number of hours per day, and minimum
    days per month, to create the output data.
    """
    SAMPLING_CONSTRAINTS = const.OBS_MIN_NUM_RESAMPLE
    APPLY_CONSTRAINTS = const.OBS_APPLY_TIME_RESAMPLE_CONSTRAINTS
    FREQS_SUPPORTED = TS_TYPE_TO_PANDAS_FREQ
    def __init__(self, input_data=None):
        self.last_setup = None
        self._input_data = None
        self.input_data = input_data
        self.valid_base_ts_types = [x for x in const.GRID_IO.TS_TYPES if 
                                    TsType(x).mulfac==1]
    
    @property
    def input_data(self):
        """Input data object that is to be resampled"""
        return self._input_data
    
    @input_data.setter
    def input_data(self, val):
        if not isinstance(val, (pd.Series, xarr.DataArray)):
            raise ValueError('Invalid input: need Series or DataArray')
        self._input_data = val
        
    @property
    def fun(self):
        """Resamplig method (depends on input data type)"""
        if isinstance(self.input_data, pd.Series):
            return resample_timeseries
        return resample_time_dataarray
    
    def _gen_idx(self, from_ts_type, to_ts_type, min_num_obs):
        if isnumeric(min_num_obs):
            return [(to_ts_type.val, int(min_num_obs))]
        if not isinstance(min_num_obs, dict):
            raise ValueError('Invalid input for min_num_obs, need dictionary '
                             'or integer, got {}'.format(min_num_obs))
        valid = self.valid_base_ts_types
        from_mul = from_ts_type.mulfac
        if from_mul != 1:
            const.print_log.warn('Ignoring multiplication factor {} in '
                                 'data with resolution {} in resampling method'
                                 .format(from_mul, from_ts_type))
        start = valid.index(from_ts_type.base)
        stop = valid.index(to_ts_type.base) 
        
        last_from = valid[start]
        idx = []
        for i in range(start+1, stop+1):
            to = valid[i]
            if to in min_num_obs and last_from in min_num_obs[to]:
                
                min_num = min_num_obs[to][last_from]
                last_from = to
                idx.append((to, min_num))
        if len(idx) == 0 or not idx[-1][0] == to_ts_type.base:
            idx.append((to_ts_type.base, 0))
        return idx
    
    def resample(self, to_ts_type, input_data=None, from_ts_type=None, 
                 how='mean', apply_constraints=False, 
                 min_num_obs=None, **kwargs):
        """Resample input data
        
        Parameters
        ----------
        input_data : pandas.Series or xarray.DataArray
            data to be resampled
        to_ts_type : str or pyaerocom.tstype.TsType
            output resolution
        how : str
            string specifying how the data is to be aggregated, default is mean
        apply_constraints : bool, optional
            if True, hierarchical resampling is applied using input 
            `samping_constraints` (if provided) or else, using constraints 
            specified in :attr:`pyaerocom.const.OBS_MIN_NUM_RESAMPLE`
        min_num_obs : dict or int, optinal
            integer or nested dictionary specifying minimum number of 
            observations required to resample from higher to lower frequency.
            For instance, if `input_data` is hourly and `to_ts_type` is
            monthly, you may specify something like::
                
                min_num_obs = 
                    {'monthly'  :   {'daily'  : 7}, 
                     'daily'    :   {'hourly' : 6}}
                    
            to require at least 6 hours per day and 7 days per month. Or, if 
            data is daily and output is monthly and   
        **kwargs
           additional input arguments passed to resampling method
          
        Returns
        -------
        pandas.Series or xarray.DataArray
            resampled data object
        """
        if not isinstance(to_ts_type, TsType):
            to_ts_type = TsType(to_ts_type)
        
        if not to_ts_type.val in self.FREQS_SUPPORTED:
            raise NotImplementedError('Cannot resample to input frequency '
                                      '{}. Choose from: {}'
                                      .format(to_ts_type, 
                                              self.FREQS_SUPPORTED.keys()))
        
        if input_data is not None:
            self.input_data = input_data
        if self.input_data is None:
            raise ValueError('Please provide data (Series or DataArray)')
        
        if apply_constraints is None:
            apply_constraints = self.APPLY_CONSTRAINTS
        
        if not apply_constraints:
            self.last_setup = dict(apply_constraints=False,
                                   min_num_obs=None)
            return self.fun(self.input_data, freq=to_ts_type.val, 
                            how=how, **kwargs)
        elif from_ts_type is None:
            const.print_log.warn('Cannot apply time resampling constraints, '
                                 'since input from_ts_type is None. Applying '
                                 'resampling to {} without any constraints'
                                 .format(to_ts_type))
            self.last_setup = dict(apply_constraints=False,
                                   min_num_obs=None)
            return self.fun(self.input_data, freq=to_ts_type.val, 
                            how=how, **kwargs)
        
        if isinstance(from_ts_type, str):
            from_ts_type = TsType(from_ts_type)
        
        if not isinstance(from_ts_type, TsType):
            raise ValueError('Invalid input for from_ts_type: {}. Need valid '
                             'str or TsType. Input arg from_ts_type is '
                             'required if resampling using hierarchical '
                             'constraints (arg apply_constraints) is activated'
                             .format(from_ts_type))
        
        if to_ts_type > from_ts_type:
            raise TemporalResolutionError('Cannot resample time-series from {} '
                                          'to {}'
                                          .format(from_ts_type, to_ts_type))
        elif to_ts_type == from_ts_type:
            const.logger.info('Input time frequency equals current frequency '
                              'of data, ignoring any resampling constraints')
            self.last_setup = dict(apply_constraints=False,
                                   min_num_obs=None)
            return self.fun(self.input_data, freq=to_ts_type.val, how=how, 
                            **kwargs)
            
        if min_num_obs is None:
            min_num_obs = self.SAMPLING_CONSTRAINTS
            
        _idx = self._gen_idx(from_ts_type, to_ts_type, min_num_obs)
        data = self.input_data
        for to_ts_type, mno in _idx:
            data = self.fun(data, freq=to_ts_type, how=how, 
                            min_num_obs=mno)
        self.last_setup = dict(apply_constraints=True,
                               min_num_obs=min_num_obs)   
        return data
    
    
if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import pyaerocom as pya

    plt.close('all')
    data = pya.io.ReadUngridded().read('EBASMC', 'scatc550dryaer')
    
    data.set_flags_nan()
    data = data.apply_filters(data_level=2, remove_outliers=True)
        
    stat = data.to_station_data('Bondvil*')
    
    hourly = stat.scatc550dryaer
    
    resampler = TimeResampler(hourly)
    
    monthly = resampler.resample('monthly', apply_constraints=True, 
                                from_ts_type='hourly')
    
    monthly1 = resampler.resample('monthly', apply_constraints=False)
    
    yearly = resampler.resample('yearly', apply_constraints=True, 
                                from_ts_type='hourly')
    
    yearly1 = resampler.resample('yearly', apply_constraints=False)
    
    fig, ax = plt.subplots(1,1, figsize=(18, 12))
    
    hourly.plot(style=' x', color='#e6e6e6', label='hourly (raw)', ax=ax)
    monthly.plot(style='-+b', label='monthly (C)')
    monthly1.plot(style=' xr', label='monthly (raw)')
    
    yearly.plot(style='-o', markersize=12, color='lime', label='yearly (C)')
    yearly1.plot(style=' o', markersize=8, color='red', label='yearly (raw)')
    
    ax.legend()
    
    fig, ax = plt.subplots(1,1, figsize=(12, 12))
    ax.plot(yearly, yearly1, ' x', color='lime')
    ax.set_xlabel('yearly (C)')
    ax.set_ylabel('yearly (raw)')



        
        
        
        
        
        
                
            
        
            
        
        
        
            
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 27 10:36:24 2019

@author: jonasg
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.cm import get_cmap
from matplotlib.colors import Normalize
from collections import OrderedDict as od
from pyaerocom.exceptions import DataCoverageError, TemporalResolutionError
from scipy.stats import kendalltau
from scipy.stats.mstats import theilslopes
import pandas as pd
from pyaerocom import const
from pyaerocom.metastandards import StationMetaData

#SEASONS = ['spring','summer','autumn','winter','all']

SEASONS = {'spring'     : [3,4,5],
           'summer'     : [6,7,8],
           'autumn'     : [9,10,11],
           'winter'     : [12, 1, 2]}

SEASON_CODES = {'spring'     : 'MAM',
                'summer'     : 'JJA',
                'autumn'     : 'SON',
                'winter'     : 'DJF'}

def _init_trends_result_dict(start_yr):
    keys = ['pval', 'm', 'm_err', 
            'n', 'y_mean', 'y_min', 'y_max', 'coverage',
            'slp', 'slp_err', 'reg0', 't0', # data specific
            'slp_{}'.format(start_yr), # period specific
            'slp_{}_err'.format(start_yr), # period specific
            'reg0_{}'.format(start_yr) # period specific
            ]
    return dict.fromkeys(keys)

def _compute_trend_error(m, m_err, v0, v0_err):
    """Computes error of trend estimate using gaussian error propagation
    
    The (normalised) trend is computed as T = m / v0
    
    where m denotes the slope of a regression line and v0 denotes the 
    normalistation value. This method computes the uncertainty of T (delta_T)
    using gaussian error propagation of uncertainties accompanying m and v0.
    
    Parameters
    ----------
    m : float
        slope in units of <U> yr-1 (where <U> denotes the unit of the data).
        (m -> "montant").
    m_err : float
        slope error (same unit as `m`)
    v0 : float
        normalisation value in units of <U>
    v0_err : float
        error of `v0` (same units as `v0`)
    
    Returns
    -------
    float
        error of T in computed using gaussian error propagation of trend 
        formula in units of %/yr
    """
    
    delta_sl = m_err / v0
    delta_ref = m * v0_err / v0**2
    return np.sqrt(delta_sl**2 + delta_ref**2) * 100

def _get_season(m, yr):
    for seas, months in SEASONS.items():
        if m in months:
            return '{}-{}'.format(seas, int(yr))
    raise ValueError('Failed to retrieve season for m={}, yr={}'
                     .format(m, yr))
 
def _mid_season(seas, yr):
    if seas=='spring':
        return np.datetime64('{}-04-15'.format(yr))    
    if seas=='summer':
        return np.datetime64('{}-07-15'.format(yr))
    if seas=='autumn':
        return np.datetime64('{}-10-15'.format(yr))    
    if seas=='winter':
        return np.datetime64('{}-01-15'.format(yr))    
    if seas=='all':
        return np.datetime64('{}-06-15'.format(yr))    
    raise ValueError('Invalid input for season (seas):', seas)
    

def _find_area(lat, lon, regions_dict=None):
    """Find area corresponding to input lat/lon coordinate
    
    Parameters
    ----------
    lat : float
        latitude
    lon : float
        longitude
    
    Returns
    -------
    str
        name of region
    """
    from pyaerocom.region import find_closest_region_coord
    reg = find_closest_region_coord(lat, lon)
    if regions_dict is not None and reg in regions_dict:
        return regions_dict[reg]
    return reg

def _years_from_periodstr(period):
    """Convert period str to start / stop years
    
    Parameters
    ----------
    period : str
        period str, e.g. '1990-2010'
        
    Returns
    -------
    int
        start year
    int 
        stop year
    """
    return [int(x) for x in period.split('-')] 

def _start_stop_period(period):
    """Convert period str to start / stop dates
    
    Parameters
    ----------
    period : str
        period str, e.g. '1990-2010'
        
    Returns
    -------
    date
        start datetime
    date
        stop datetime
    """
    from datetime import date
    y0, y1, = _years_from_periodstr(period)
    return (date(y0, 1, 1), date(y1, 12, 31))

class TrendsEngine(object):
    """Result object for trends analysis
    
    Parameters
    ----------
    var_name : str
        name of variable 
        
    Attributes
    ----------
    var_name : str
        name of variable 
    meta : dict
        dictionary containing relevant meta information about data
    daily : pandas.Series
        daily data
    monthly : pandas.Series
        monthly data
    yearly : dict
        dictionary containing yearly data for each season. Keys are season names, 
        values are instances of :class:`pandas.Series`
    results : dict
        Nested dictionary containing results from trends analysis. First layer
        is season, second layer is period, and values are dictionaries with 
        results (cf. keys of dictionary returned by 
        :func:`_init_trends_result_dict`)
    """
    CMAP = get_cmap('bwr')
    NORM = Normalize(-10, 10)
    
    def __init__(self, var_name=None, **kwargs):
        self.var_name = var_name
        self.meta = StationMetaData()
        
        self._daily = None
        self._monthly = None
        self.yearly = od()
        
        self.results = od()
        self._mobs = None 
        
    @property
    def daily(self):
        """Daily timeseries"""
        return self._daily
    
    @daily.setter
    def daily(self, val):
        if not isinstance(val, pd.Series):
            raise ValueError('Invalid input for attr. daily. Need '
                             'pandas.Series.')
        self._daily = val
        self.compute_monthly()
    
    @property
    def monthly(self):
        """Monthly timeseries"""
        return self._monthly
    
    @monthly.setter
    def monthly(self, val):
        if not isinstance(val, pd.Series):
            raise ValueError('Invalid input for attr. daily. Need '
                             'pandas.Series.')
        self._monthly = val
        self._mobs = _make_mobs_dataframe(val)
        
    @property
    def has_daily(self):
        """Boolean specifying whether daily data is available"""
        return True if isinstance(self.daily, pd.Series) else False
    
    @property
    def has_monthly(self):
        """Boolean specifying whether monthly data is available"""
        return True if isinstance(self.monthly, pd.Series) else False
    
    def compute_monthly(self, daily=None, **kwargs):
        """Computes monthly timeseries from daily
        
        Note
        ----
        Requires that daily timeseries data is available in :attr:`daily` or
        provided via **kwargs using keyword `daily`. This method also computes
        the monthly dataframe that contains the seasons and is available via
        :attr:`_mobs`.
        
        Parameters
        ----------
        daily : pandas.Series
            daily timeseries (will overwrite :attr:`daily`)
        **kwargs
            input args passed to :func:`pyaerocom.helpers.resample_timeseries`
            such as (`min_num_obs`)
        
        Returns
        -------
        pandas.Series
            monthly timeseries
        """
        from pyaerocom.helpers import resample_timeseries
        if daily is not None:
            self.daily = daily
        monthly = resample_timeseries(self.daily, freq='monthly', **kwargs)
        mobs = _make_mobs_dataframe(monthly)
        self._mobs = mobs
        self._monthly = monthly
        return monthly
    
    @property
    def seasons_avail(self):
        """List of all seasons for which trends are available"""
        return self.results.keys()
    
    @property
    def periods_avail(self):
        """List of all periods for which trends are available"""
        periods = []
        for seas, period_data in self.results.items():
            for per in list(period_data):
                if not per in periods:
                    periods.append(per)
        return periods
        
    def get_yearly(self, season):
        """Get yearly time series for a certain season
        
        Parameters
        ----------
        season : str
            name of season
        
        Returns
        -------
        pandas.Series
            yearly data for that season
        
        Raises
        ------
        AttributeError
            if yearly data is not available for that season
        """
        if not season in self.yearly:
            raise AttributeError('Yearly data for season {} is not available'
                                 .format(season))
        return self.yearly[season]
    
    def _get_trend_data(self, season, period):
        if not self.has_monthly:
            raise AttributeError('No monthly data available')
        elif not season in self.seasons_avail:
            raise AttributeError('No results available for season {}'.format(season))
        elif not period in self.periods_avail:
            raise AttributeError('No results available for period {}'.format(period))
        
        result = self.results[season][period]
        if result['m'] is None:
            raise AttributeError('No slope information available')
        
        start_data = self.monthly.index.year[0]
        stop_data = self.monthly.index.year[-1]
        
        start_period, stop_period = _years_from_periodstr(period)
        
        
        (_,_, idx_data, num_dates_data) = _init_period_dates(start_data, 
                                                             stop_data, 
                                                             season)
        (_,_, idx_period, num_dates_period) = _init_period_dates(start_period, 
                                                                 stop_period, 
                                                                 season)
        if start_data < start_period:
            start_data = start_period
        
        
        regr_data = result['m'] * num_dates_data + result['yoffs']
        regr_period = result['m'] * num_dates_period + result['yoffs']
        s_data = pd.Series(regr_data, idx_data)
        s_period = pd.Series(regr_period, idx_period)
        try:
            td = result['slp']
            td_err = result['slp_err']
        except:
            td = None
            td_err = None
        try:
            tp = result['slp_{}'.format(start_period)]
            tp_err = result['slp_{}_err'.format(start_period)]
        except:
            tp = None
            tp_err = None
        pval = result['pval']
        tdstr = ''
        tpstr = ''
        try:
            tdstr = (r'$\mathcal{{T}}_{{{}}}: {:.2f}\,\pm\,{:.2f}\,\%/yr$'
                     .format(start_data, td, td_err))
        except: 
            pass
        try:
            tpstr = (r'$\mathcal{{T}}_{{{}}}: \mathbf{{{:.2f}\,\pm\,{:.2f}\,\%/yr}}$'
                     .format(start_period, tp, tp_err))
        except:
            pass
        try:
            tpstr += '; pval: {:.1e}'.format(pval)
        except:
            tdstr += '; pval: {:.1e}'.format(pval)
        return (s_data, s_period, td, tp, tdstr, tpstr)
    
    def compute_trend(self, start_year, stop_year, season=None, 
                      slope_confidence=.68):
        if slope_confidence is None:
            slope_confidence = .68
        if self._mobs is None:
            raise ValueError('Cannot compute trends: monthly data is not '
                             'available')
        mobs = self._mobs
        start_year, stop_year, period_str, yrs = _init_period(mobs, start_year, 
                                                              stop_year)
        
        if season in [None, 'all']:
            seas = 'all'
        elif season in SEASONS:
            seas = season
           
        if not 'seas' in self.yearly:
            self['yearly'][seas] = yearly = _get_yearly(mobs, seas, yrs)
        else:
            yearly = self['yearly'][seas]
        
        dates = yearly.index.values
        values = yearly.values 
        (start_date, 
         stop_date, 
         period_index, 
         num_dates_period) = _init_period_dates(start_year, stop_year, seas)
        
        # get period filter mask
        tmask = np.logical_and(dates>=start_date, 
                               dates<=stop_date) 
            
        # apply period mask to jsdate vector and value vector
        dates_data = dates[tmask]
        
        # vector containing data values
        vals = values[tmask]
        
        valid = ~np.isnan(vals)
        
        #works only on not nan values
        dates_data = dates_data[valid]
        vals = vals[valid]
        
        num_dates_data = dates_data.astype('datetime64[Y]').astype(np.float64)
        
        # create empty dictionary that is used to store trends results
        result = _init_trends_result_dict(start_year)
        
        #TODO: len(y) is number of years - 1 due to midseason averages
        result['n'] = len(vals)
        
        if len(vals) > 2:
            result['y_mean'] = np.nanmean(vals)
            result['y_min'] = np.nanmin(vals)
            result['y_max'] = np.nanmax(vals)
            
            #Mann / Kendall test
            [tau, pval] = kendalltau(x=num_dates_data, y=vals)
            
            
            (slope, 
             yoffs, 
             slope_low, 
             slope_up) = theilslopes(y=vals, x=num_dates_data, 
                                     alpha=slope_confidence)
            
            # estimate error of slope at input confidence level
            slope_err = np.mean([abs(slope - slope_low), 
                                 abs(slope - slope_up)])
            
            reg_data = slope * num_dates_data + yoffs
            reg_period = slope * num_dates_period  + yoffs
            
            
            # value used for normalisation of slope to compute trend T
            # T=m / v0
            v0_data = reg_data[0]
            v0_period = reg_period[0]
            
            # Compute the mean residual value, which is used to estimate
            # the uncertainty in the normalisation value used to compute
            # trend
            mean_residual = np.mean(np.abs(vals - reg_data))
            
            # trend is slope normalised by first reference value. 
            # 2 trends are computed, 1. the trend using the first value of
            # the regression line at the first available data year, 2. the
            # trend corresponding to the value corresponding to the first
            # year of the considered period.
            
            trend_data = slope / v0_data * 100
            trend_period =  slope / v0_period * 100
            
            # Compute errors of normalisation values
            v0_err_data = mean_residual
            t0_data, tN_data = num_dates_data[0], num_dates_data[-1]
            t0_period = num_dates_period[0]
            
            # sanity check
            assert t0_data < tN_data
            assert t0_period <= t0_data
            
            dt_ratio = (t0_data - t0_period) / (tN_data - t0_data)
            
            v0_err_period = v0_err_data * (1 + dt_ratio)
            
            trend_data_err = _compute_trend_error(m=slope,
                                                  m_err=slope_err,
                                                  v0=v0_data,
                                                  v0_err=v0_err_data)
                                                      
            trend_period_err = _compute_trend_error(m=slope,
                                                    m_err=slope_err,
                                                    v0=v0_period,
                                                    v0_err=v0_err_period)
                            
            result['pval'] = pval
            result['m'] = slope
            result['m_err'] =slope_err
            result['yoffs'] = yoffs
            
            result['slp'] = trend_data
            result['slp_err'] = trend_data_err
            result['reg0'] = v0_data
            tp, tperr, v0p = None, None, None
            if v0_period > 0:
                tp = trend_period
                tperr = trend_period_err
                v0p = v0_period
            result['slp_{}'.format(start_year)] = tp
            result['slp_{}_err'.format(start_year)] = tperr
            result['reg0_{}'.format(start_year)] = v0p
            result['period'] = period_str
    
        if not seas in self.results:
            self.results[seas] = od()
        self.results[seas][period_str] = result

        return result

    def to_json(self):
        raise NotImplementedError
    
    def from_json(self, filepath):
        raise NotImplementedError
    
    def _plot_trend_result(self, tr, ax):
        pass
    
    def get_trend_color(self, trend_val):
        return self.CMAP(self.NORM(trend_val))
    
    def plot(self, season='all', period=None, ax=None):
        if not season in self.seasons_avail:
            raise AttributeError('No results available for season {}'
                                 .format(season))
        if period is None:
            if len(self.results[season]) > 1:
                raise ValueError('Found multiple trends for different periods: '
                                 '{}. Please specify period...'
                                 .format(list(self.results[season].keys())))
            period = list(self.results[season].keys())[0]
        if ax is None:
            fig, ax = plt.subplots(1,1, figsize=(18, 8))
        if self.has_daily:
            ax.plot(self.daily,'-', marker ='.',  label='daily', c='#d9d9d9')
        if self.has_monthly:
            ax.plot(self.monthly, label='monthly', c='#4d4d4d')
        ax.plot(self.get_yearly(season), ' ok', label='yearly')
        if period in self.periods_avail:
            
            (s_data, s_period, td, tp, tdstr,
             tpstr)= self._get_trend_data(season, period)
            
            ax.plot(s_data, '-', color=self.get_trend_color(td), label='trend', 
                    lw=2)
            ax.plot(s_period, '--', color=self.get_trend_color(tp))
            ax.text(0.01, 0.95, tpstr, transform=ax.transAxes,
                    fontsize=14)
            ax.text(0.01, 0.9, tdstr, transform=ax.transAxes, 
                    fontsize=14)
            
        ax.yaxis.grid(c='#d9d9d9', ls='-.')
        ylbl = self.var_name
        if self.var_name is not None and 'units' in self.meta:
            u = str(self.meta['units'])
            if not u in ['', '1']:
                ylbl += ' [{}]'.format(u)
        ax.set_ylabel(ylbl)
        tit = '' 
        if self.meta['station_name'] is not None:
            tit += self.meta['station_name']
            try:
                self.meta.load_dataset_info()
            except:
                pass
            dsinfo = self.meta.dataset_str()
            if dsinfo is not None:
                tit += '; {}'.format(dsinfo)
            tit += ' - '
        tit += period
        ax.set_title(tit)
        ax.legend(loc=1)
    
        ax.set_xlim(_start_stop_period(period))
        
        return ax
    
    def __getitem__(self, key):
        return self.__getattribute__(key)
        
    def __setitem__(self, key, val):
        self.__setattr__(key, val)
    
def _make_mobs_dataframe(monthly_series):
    
    
    dates = monthly_series.index

    #dates = pd.DatetimeIndex(index)
    d = dict(date=dates,
             value=monthly_series.values,
             year=dates.year,
             month=dates.month,
             day=dates.day)    

    mobs = pd.DataFrame(d)
    
    add_season = lambda row: _get_season(row['month'], row['year'])
    # add season column to dataframe
    mobs['season'] = mobs.apply(add_season, axis=1)    
    
    # remove all NaN values
    mobs = mobs.dropna(subset=['value'])
    
    return mobs

def _init_period(mobs, start_year=None, stop_year=None):
    yrs = sorted(np.unique(mobs['year']))
    if start_year is None:
        start_year = yrs[0]
    if stop_year is None:
        stop_year =  yrs[-1]
    if stop_year < yrs[0] or start_year > yrs[-1]:
        from pyaerocom.exceptions import TimeMatchError
        raise TimeMatchError('Input period {}-{} is not covered by data ({}-{})'
                             .format(start_year, stop_year, yrs[0],yrs[-1]))
    pstr = '{}-{}'.format(int(start_year), int(stop_year))
    return (start_year, stop_year, pstr, yrs)

def _get_yearly(mobs, seas, yrs):
    dates = []
    values = []
    for yr in yrs:
        dates.append(_mid_season(seas, yr))
        if seas == 'all': #yearly trends
            subset = mobs[mobs['year'] == yr]
        else:
            subset = mobs[mobs['season'].str.contains('{}-{}'.format(seas, yr))]
        #needs 4 seasons to compute seasonal average to avoid biases
        if seas=='all' and len(np.unique(subset['season'].values)) < 4:
            values.append(np.nan)
        else:
            values.append(np.nanmean(subset['value']))

    return pd.Series(values, index=dates)

def _init_period_dates(start_year, stop_year, season):
    start_date = _mid_season(season, start_year)
    stop_date = _mid_season(season, stop_year)
        
    # datetime index covering whole input period (data may not be fully covered
    # in the whole period)
    period_index = pd.date_range(start=start_date,
                                 end=stop_date, 
                                 freq=pd.DateOffset(years=1))
    
    num_dates_period = period_index.values.astype('datetime64[Y]').astype(np.float64)
    return (start_date, stop_date, period_index, num_dates_period)

def compute_trends_station(station, var_name, start_year=None, 
                           stop_year=None, season=None, slope_confidence=0.68,
                           **alt_range):
    # load additional information about data source (if applicable)
    if not 'trends' in station:
        station['trends'] = od()
    tr = station['trends']
    if not var_name in tr:
        station['trends'][var_name] = trv = TrendsEngine(var_name)
    else:
        trv = station['trends'][var_name]

    freq = station.get_var_ts_type(var_name)
    
    ts_types = const.GRID_IO.TS_TYPES
    
    if not trv.has_daily:
        if not freq in ts_types or (ts_types.index(freq) <= ts_types.index('daily')):
            trv['daily'] = station.to_timeseries(var_name, freq='daily', **alt_range)
    # monthly is mandatory
    if not trv.has_monthly:
        if freq in ts_types and ts_types.index(freq) >= ts_types.index('monthly'):
            raise TemporalResolutionError('Need monthly or higher')
        ms = station.to_timeseries(var_name, freq='monthly', **alt_range)
        trv['monthly'] = ms
    else:
        ms = trv['monthly']
        
    if len(ms) == 0 or all(np.isnan(ms)):
        raise DataCoverageError('Failed to retrieve monthly timeseries for '
                                '{} ({})'.format(station.station_name,
                                 var_name))
      
    if trv._mobs is None:
        trv._mobs = _make_mobs_dataframe(ms)
    
    result = trv.compute_trend(start_year, stop_year, season, 
                               slope_confidence)
    
    
    trv.meta.update(station.get_meta(add_none_vals=True))
    if var_name in station.var_info:
        trv.meta.update(station.var_info[var_name])
    return result
    
if __name__ == '__main__':
    import pyaerocom as pya
    plt.close('all')
    
    r = pya.io.ReadUngridded()
    data = r.read('AeronetSunV3Lev2.daily', vars_to_retrieve='od550aer',
                  file_pattern='Solar*')
    
    stat = data.to_station_data('Solar*')
    
    compute_trends_station(stat, 'od550aer', start_year=1995, 
                           stop_year=2017)
    
    stat.compute_trend('od550aer', 2002, 2012)
    tr = stat.trends['od550aer']
    
    ax = tr.plot(season='all', period='1995-2017')
    
    ax = tr.plot(season='all', period='2002-2012')
    plt.show()
    
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul  9 14:14:29 2018
"""
import numpy.testing as npt
import numpy as np
import os
import pytest
from pyaerocom.test.settings import TEST_RTOL, lustre_unavail
from pyaerocom.io.read_aeronet_sunv3 import ReadAeronetSunV3


TEST_VARS = ['od550aer', 'ang4487aer']
    
def make_dataset():
    r = ReadAeronetSunV3()
    return r.read(file_pattern='G*', 
                  vars_to_retrieve=TEST_VARS)

@lustre_unavail
@pytest.fixture(scope='session')
def aeronetsunv3lev2_subset():
    return make_dataset()

@lustre_unavail
def test_shape_ungridded(aeronetsunv3lev2_subset):
    assert aeronetsunv3lev2_subset.shape == (63170, 12)
    
@lustre_unavail
def test_meta_blocks_ungridded(aeronetsunv3lev2_subset):
    assert len(aeronetsunv3lev2_subset.metadata) == 29
    assert len(aeronetsunv3lev2_subset.unique_station_names) == 29
    
    assert aeronetsunv3lev2_subset.unique_station_names == ['GISS',
                                         'GORDO_rest',
                                         'GOT_Seaprism',
                                         'GSFC',
                                         'Gageocho_Station',
                                         'Gainesville_Airport',
                                         'Gaithersburg',
                                         'Galata_Platform',
                                         'Gandhi_College',
                                         'Gangneung_WNU',
                                         'Georgia_Tech',
                                         'Glasgow_MO',
                                         'Gloria',
                                         'Gobabeb',
                                         'Goldstone',
                                         'Gorongosa',
                                         'Gosan_SNU',
                                         'Gotland',
                                         'Gozo',
                                         'Graciosa',
                                         'Granada',
                                         'Grand_Forks',
                                         'Granite_Island',
                                         'Grizzly_Bay',
                                         'Guadeloup',
                                         'Gual_Pahari',
                                         'Guam',
                                         'Gustav_Dalen_Tower',
                                         'Gwangju_GIST']

@lustre_unavail
def test_od550aer_meanval_stats(aeronetsunv3lev2_subset):
    no_odcount = 0
    mean_vals = []
    std_vals = []
    for stat in aeronetsunv3lev2_subset:
        if not 'od550aer' in stat:
            no_odcount += 1
            continue
        mean = np.mean(stat.od550aer)
        if np.isnan(mean):
            no_odcount += 1
            continue
        mean_vals.append(mean)
        std_vals.append(np.std(stat.od550aer))
    assert no_odcount == 5
    npt.assert_allclose(actual=[np.mean(mean_vals), 
                                np.mean(std_vals)],
                        desired=[0.209, 0.149], atol=1e-2)
    
@lustre_unavail
def test_ang4487aer_meanval_stats(aeronetsunv3lev2_subset):
    no_odcount = 0
    mean_vals = []
    std_vals = []
    for stat in aeronetsunv3lev2_subset:
        if not 'ang4487aer' in stat:
            no_odcount += 1
            continue
        mean = np.mean(stat.ang4487aer)
        if np.isnan(mean):
            no_odcount += 1
            continue
        mean_vals.append(mean)
        std_vals.append(np.std(stat.ang4487aer))
    assert no_odcount == 1
    npt.assert_allclose(actual=[np.mean(mean_vals), 
                                np.mean(std_vals)],
                        desired=[1.202, 0.337], atol=1e-2)
    
@lustre_unavail
def test_load_berlin():
    dataset = ReadAeronetSunV3()
    files = dataset.find_in_file_list('*Berlin*')
    assert len(files) == 1
    assert os.path.basename(files[0]) == 'Berlin_FUB.lev30'
    data = dataset.read_file(files[0],
                             vars_to_retrieve=['od550aer'])
    
    test_vars = ['od440aer',
                 'od500aer',
                 'od550aer',
                 'ang4487aer']
    assert all([x in data for x in test_vars])
    
    # more than 100 timestamps
    assert all([len(data[x]) > 100 for x in test_vars])
    
    assert isinstance(data['dtime'][0], np.datetime64)
    t0 = data['dtime'][0]
    
    assert t0 == np.datetime64('2014-07-06T12:00:00')
    
    
    first_vals = [data[var][0] for var in test_vars]
    
    nominal = [0.224297, 0.178662, 0.148119, 1.967039]
    npt.assert_allclose(actual=first_vals, desired=nominal, rtol=TEST_RTOL)
    
    
    
    
if __name__=="__main__":
    
    test_load_berlin()
    aeronetsunv3lev2_subset = make_dataset()
    test_shape_ungridded(aeronetsunv3lev2_subset)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module contains functionality related to regions in pyaerocom
"""
from os.path import join, exists
from ast import literal_eval
from collections import OrderedDict as od
from configparser import ConfigParser
from pyaerocom import __dir__
from pyaerocom._lowlevel_helpers import BrowseDict

class Region(BrowseDict):
    """Interface that specifies an AEROCOM region
    
    Note
    ----
    Extended dictionary-like object
    
    Attributes
    ----------
    name : str
        ID of region (e.g. "EUROPE")
    lon_range : :obj:`list` or :obj:`tuple`
        longitude range (min, max) covered by region
    lat_range : :obj:`list` or :obj:`tuple`
        latitude range (min, max) covered by region
    lon_range_plot : :obj:`list` or :obj:`tuple`
        longitude range (min, max) used for plotting region. Defaults to 
        values of :attr:`lon_range` but may be extended, see e.g.
        `the example of india <http://aerocom.met.no/DATA/SURFOBS/
        ECMWF_OSUITE_NRT/plots/OD550_AER_an2018_d20180319_INDIA_MAP.ps.png>`__
    lat_range_plot : :obj:`list` or :obj:`tuple`
        latitude range (min, max) used for plotting region. Defaults to 
        values of :attr:`lat_range` but may be extended, see e.g.
        `the example of india <http://aerocom.met.no/DATA/SURFOBS/
        ECMWF_OSUITE_NRT/plots/OD550_AER_an2018_d20180319_INDIA_MAP.ps.png>`__
    
    Parameters
    ----------
    name : str
        ID of region (e.g. "EUROPE")
    lon_range : :obj:`list` or :obj:`tuple`
        longitude range (min, max) covered by region
    lat_range : :obj:`list` or :obj:`tuple`
        latitude range (min, max) covered by region
        
    Example
    -------
    Just initiate with a valid region ID
    
        >>> from pyaerocom import Region
        >>> europe = Region("EUROPE")
        >>> india = Region("INDIA")
        >>> print(europe)
        pyaeorocom Region
        Name: EUROPE
        Longitude range: [-20, 70]
        Latitude range: [30, 80]
        Longitude range (plots): [-20, 70]
        Latitude range (plots): [30, 80]
        >>> print(india)
        pyaeorocom Region
        Name: INDIA
        Longitude range: [65, 90]
        Latitude range: [5, 35]
        Longitude range (plots): [50, 100]
        Latitude range (plots): [0, 40]
        
    """
    def __init__(self, name="WORLD", lon_range=None, lat_range=None, 
                 **kwargs):
        self._name = name
        # longitude / latitude range of data 
        self.lon_range = lon_range
        self.lat_range = lat_range
        
        # longitude / latitude range of data in plots
        self.lon_range_plot = None
        self.lat_range_plot = None
        
        self.lon_ticks = None
        self.lat_ticks = None
        
        if isinstance(name, str) and any([x == None for x in [lon_range, lat_range]]):
           self.import_default(name) 
        
        for k, v in kwargs.items():
            if k in self:
                self[k] = v
    
    @property
    def name(self):
        """Name of region"""
        return self._name
    
    def import_default(self, name):
        """Import information about default region
        
        Parameters
        ----------
        name : str
            strind ID of region (must be specified in `regions.ini <https://
            github.com/metno/pyaerocom/blob/master/pyaerocom/data/regions.ini>`__ 
            file)
            
        Raises
        ------
        IOError
            if regions.ini file does not exist
        NameError
            if default region with ID specified by input parameter ``name`` 
            cannot be found in regions.ini file
        """
        name = name.upper()
        fpath = join(__dir__, "data", "regions.ini")
        if not exists(fpath):
            raise IOError("Regions ini file could not be found: %s"
                          %fpath)
        conf_reader = ConfigParser()
        conf_reader.read(fpath)
        if not name in conf_reader:
            raise AttributeError("No default region available for %s" %name)
        self._name = name
        for key, val in conf_reader[name].items():
            if key in self.keys():
                if "," in val:
                    #list of values
                    val = list(literal_eval(val))#[float(x) for x in val.split(",")]
                else:
                    try:
                        val = int(val)
                    except:
                        pass
                self[key] = val
        if self.lon_range_plot is None:
            self.lon_range_plot = self.lon_range
        if self.lat_range_plot is None:
            self.lat_range_plot = self.lat_range
    
    @property
    def center_coordinate(self):
        """Center coordinate of this region"""
        latc = self.lat_range[0] + (self.lat_range[1] - self.lat_range[0])/2
        lonc = self.lon_range[0] + (self.lon_range[1] - self.lon_range[0])/2
        return (latc, lonc)
    
    def distance_to_center(self, lat, lon):
        """Compute distance of input coordinate to center of this region
        
        Parameters
        ----------
        lat : float
            latitude of coordinate
        lon : float
            longitude of coordinate
        
        Returns
        -------
        float
            distance in km
        """
        from pyaerocom.geodesy import calc_distance
        cc = self.center_coordinate
        return calc_distance(lat0=cc[0], 
                             lon0=cc[1], 
                             lat1=lat, 
                             lon1=lon)
        
    def contains_coordinate(self, lat, lon):
        """Check if input lat/lon coordinate is contained in region
        
        Parameters
        ----------
        lat : float
            latitude of coordinate
        lon : float
            longitude of coordinate
        
        
        Returns
        -------
        bool
            True if coordinate is contained in this region, False if not
        """
        lat_ok = self.lat_range[0] <= lat <= self.lat_range[1]
        lon_ok = self.lon_range[0] <= lon <= self.lon_range[1]
        return lat_ok * lon_ok
    
    def __contains__(self, val):
        if not isinstance(val, tuple):
            raise TypeError('Invalid input, need tuple')
        if not len(val) == 2:
            raise ValueError('Invalid input: coordinate must contain 2 '
                             'elements (lat, lon)')
        return self.contains_coordinate(lat=val[0], lon=val[1])
        
    def __repr__(self):
       return ("Region %s %s" %(self.name, super(Region, self).__repr__()))
   
    def __str__(self):
        s = ("pyaeorocom Region\nName: %s\n"
             "Longitude range: %s\n"
             "Latitude range: %s\n" 
             "Longitude range (plots): %s\n"
             "Latitude range (plots): %s" 
             %(self.name, self.lon_range, self.lat_range,
               self.lon_range_plot, self.lat_range_plot))
        return s

def all():
    """Wrapper for :func:`get_all_default_region_ids`"""
    return get_all_default_region_ids()

def get_all_default_region_ids():
    """Get list containing IDs of all default regions
    
    Returns
    -------
    list
        all default region IDs (sections in `regions.ini <https://github.com/
        metno/pyaerocom/blob/master/pyaerocom/data/regions.ini>`__ file)
    """
    fpath = join(__dir__, "data", "regions.ini")
    if not exists(fpath):
        raise IOError("File conventions ini file could not be found: %s"
                      %fpath)
    conf_reader = ConfigParser()
    conf_reader.read(fpath)
    all_ids = []
    for rid, region in conf_reader.items():
        if not rid == "DEFAULT":
            all_ids.append(rid)
        
    return all_ids

def get_all_default_regions():
    """Get dictionary containing all default regions from region.ini file
    
    Note
    ----
        The values are already :class:`Region` instances, use 
        :func:`get_all_default_region_ids`
    
    Returns
    -------
    dict
        dictionary containing all default regions that are found in 
        the `regions.ini <https://github.com/metno/pyaerocom/blob/
        master/pyaerocom/data/regions.ini>`__ file
        
    """
    fpath = join(__dir__, "data", "regions.ini")
    if not exists(fpath):
        raise IOError("File conventions ini file could not be found: %s"
                      %fpath)
    conf_reader = ConfigParser()
    conf_reader.read(fpath)
    all_regions = od()
    for region in conf_reader:
        if not region == "DEFAULT":
            all_regions[region] = Region(region)
        
    return all_regions
   
def get_regions_coord(lat, lon, **add_regions):
    """Get all regions that contain input coordinate
    
    Parameters
    ----------
    lat : float
        latitude of coordinate
    lon : float
        longitude of coordinate
    **add_regions
        additional regions other than default regions, that are supposed to 
        be considered
        
    Returns
    -------
    list
        list of regions that contain this coordinate
    """
    regs = []
    all_regs = get_all_default_regions()
    for an, ar in add_regions.items():
        if isinstance(ar, Region):
            all_regs[an] = ar

    for rname, reg in all_regs.items():
        if rname == 'WORLD':
            continue
        if reg.contains_coordinate(lat, lon):
            regs.append(rname)
    if len(regs) == 0:
        regs.append('WORLD')
    return regs

def find_closest_region_coord(lat, lon, **add_regions):
    """Find region that has it's center closest to input coordinate
    
    Parameters
    ----------
    lat : float
        latitude of coordinate
    lon : float
        longitude of coordinate
    **add_regions
        additional regions other than default regions, that are supposed to 
        be considered
        
    Returns
    -------
    str
        name of region
    """
    regs = get_regions_coord(lat, lon, **add_regions)
    
    if len(regs) == 1:
        return regs[0]
    else:
        default_ids = get_all_default_region_ids()
        min_dist = 1e6
        best = None
        for rname in regs:
            if rname in default_ids:
                r = Region(rname)
            else:
                r = add_regions[rname]
            d = r.distance_to_center(lat, lon)
            if d < min_dist:
                min_dist = d
                best=rname
        return best

def valid_region(name):
    """Boolean specifying whether input region is valid or not"""
    if isinstance(name, str):
        return True if name.upper() in get_all_default_region_ids() else False
    return False

if __name__=="__main__":

    r = Region()
    r.import_default("EUROPE")
        
    all_ids = get_all_default_region_ids()
    
    lat, lon = 89, 30
    reg = find_closest_region_coord(lat, lon)
    print(reg)
        
    print(valid_region('europe'))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 12 14:45:43 2018

@author: jonasg
"""

import pytest
import numpy.testing as npt
from datetime import datetime
from pyaerocom.test.settings import TEST_RTOL, lustre_unavail
from pyaerocom import GriddedData

### Helpers that may be used in __main__ for testing
def _load_osuite():
    from pyaerocom.io.testfiles import get
    test_file = get()['models']['ecmwf_osuite']
    return GriddedData(test_file, var_name="od550aer")

def _load_cams_rean():
    from pyaerocom.io import ReadGridded
    r = ReadGridded(data_id="ECMWF_CAMS_REAN")
    return r.read_var('od550aer', ts_type='daily',
                      start=2010, stop=2013)
    
### fixtures
@lustre_unavail
@pytest.fixture(scope='module')
def data_cci():
    '''import example data from Aerosol CCI
    
    The fixture property makes sure that this "variable" is only created once
    for the entire scope of this test session within this module
    '''
    from pyaerocom.io.testfiles import get
    test_file = get()['models']['aatsr_su_v4.3']
    return GriddedData(test_file, var_name="od550aer")

@lustre_unavail
@pytest.fixture(scope='module')
def data_cams_rean():
    return _load_cams_rean()

@lustre_unavail
@pytest.fixture(scope='module')
def data_osuite():
    '''import example data from ECMWF_OSUITE
    
    The fixture property makes sure that this "variable" is only created once
    for the entire scope of this test session within this module
    '''
    return _load_osuite()    

### tests
@lustre_unavail
def test_longitude(data_cci, data_osuite):
    """Test if longitudes are defined right"""
    lons_cci = data_cci.longitude.points
    lons_osuite = data_osuite.longitude.points
    nominal = [-179.5, 179.5, -180.0, 179.6]
    vals = [lons_cci.min(), lons_cci.max(),
            lons_osuite.min(), lons_osuite.max()]
    npt.assert_allclose(actual=vals, desired=nominal, rtol=TEST_RTOL)
    
@lustre_unavail
def test_latitude(data_cci):
    """test latitude array"""
    nominal_eq = ['arc_degree', 0]
    vals_eq = [data_cci.latitude.units.name,
               int(sum(data_cci.latitude.points))]
    npt.assert_array_equal(nominal_eq, vals_eq)

@lustre_unavail    
def test_time(data_cci, data_osuite):
    """Test time dimension access and values"""
    time_cci = data_cci.time
    time_osuite = data_osuite.time
    nominal_eq = ["gregorian", 
                  "julian",
                  'day since 2018-01-01 00:00:00.00000000 UTC',
                  'day since 2008-01-01 00:00:00.00000000 UTC', 
                  True, 
                  False]
    vals_eq = [time_osuite.units.calendar, 
               time_cci.units.calendar,
               time_osuite.units.name, 
               time_cci.units.name,
               isinstance(time_osuite.cell(0).point, datetime),
               isinstance(time_cci.cell(0).point, datetime)]
    npt.assert_array_equal(nominal_eq, vals_eq)


@lustre_unavail
def test_downscale_time(data_cams_rean):
    
    data = data_cams_rean
    print(data)
    
    monthly = data.downscale_time('monthly')
    yearly = data.downscale_time('yearly')
    
    npt.assert_array_equal(data.shape, (1097, 161, 320))
    npt.assert_array_equal(monthly.shape, (37, 161, 320))
    npt.assert_array_equal(yearly.shape, (4, 161, 320))
    
    mean_vals = [data.mean(), monthly.mean(), yearly.mean()]
    npt.assert_allclose(actual=mean_vals,
                        desired=[0.1213392669166126, 
                                 0.12069144475849365, 
                                 0.11591256935171756], rtol=TEST_RTOL)

if __name__=="__main__":
    import warnings
    warnings.filterwarnings('ignore')
    pytest.main()
    
    
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 13 09:31:56 2019

@author: jonasg
"""

VAR_MAPPING = {'od550aer'       : ['AOD', '2D', 'Optical properties'],
               'od550csaer'     : ['AOD (clear sky)', '2D', 'Optical properties'],
               'od550lt1aer'    : ['AOD<1um', '2D', 'Optical properties'],
               'od550gt1aer'    : ['AOD>1um', '2D', 'Optical properties'],
               'abs550aer'      : ['AAOD', '2D', 'Optical properties'],
               'ang4487aer'     : ['AE', '2D', 'Optical properties'],
               'ang4487csaer'   : ['AE (clear sky)', '2D', 'Optical properties'],
               'scatc550dryaer' : ['Scat. coef. (dry)', '3D', 'Optical properties'],
               'scatc550aer'    : ['Scat. coef.', '3D', 'Optical properties'],
               'absc550aer'     : ['Abs. coef.', '3D', 'Optical properties'],
               'ec532aer'       : ['Ext. coeff.', '3D', 'Optical properties'],
               'bscatc532aer'   : ['Backscat. coeff.', '3D', 'Optical properties'],
               'concpm10'      : ['PM10', '3D', 'Particle concentrations'],
               'concpm25'      : ['PM2.5', '3D', 'Particle concentrations'],
               'concso4'       : ['SO4 (Aerosol)', '3D', 'Particle concentrations'],
               'concso4pr'     : ['SO4 (precip.)', '3D', 'Particle concentrations'],
               'conco3'        : ['O3', '3D', 'Gas concentrations'],
               'concso2'       : ['SO2', '3D', 'Gas concentrations'],
               }


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for geographical calculations 

This module contains low-level methods to perform geographical calculations, 
(e.g. distance between two coordinates)
"""
from pyaerocom import GEONUM_AVAILABLE
from pyaerocom import print_log, logger
import numpy as np
import os

def get_topo_data(lat0, lon0, lat1=None, lon1=None, topo_dataset='srtm', 
                  topodata_loc=None, try_etopo1=False):
    """Retrieve topographic altitude for a certain location
    
    Currently works only if :mod:`geonum` is installed. Supports topography
    datasets supported by geonum. These are currently (20 Feb. 19) srtm 
    (SRTM dataset, default, automatic access if online) and etopo1 
    (ETOPO1 dataset, lower resolution, must be available on local machine or
    server). 
    
    Parameters
    ----------
    lat0 : float 
            start longitude for data extraction
    lon0 : float 
        start latitude for data extraction
    lat1 : float 
        stop longitude for data extraction (default: None). If None only 
        data around lon0, lat0 will be extracted.
    lon1 : float
        stop latitude for data extraction (default: None). 
        If None only data around lon0, lat0 will be extracted
    topo_dataset : str
        name of topography dataset
    topodata_loc : str
        filepath or directory containing supported topographic datasets
    try_etopo1 : bool
        if True and if access fails via input arg `topo_dataset`, then try
        to access altitude using ETOPO1 dataset.
        
    Returns
    -------
    geonum.TopoData
        data object containing topography data in specified range 
    
    Raises
    ------
    ValueError
        if altitude data cannot be accessed
    """
    if not GEONUM_AVAILABLE:
        raise ModuleNotFoundError('Feature disabled: geonum library is not '
                                  'installed')
    import geonum
    if topodata_loc is None:
        from pyaerocom import const
        if topo_dataset in const.SUPPLDIRS and os.path.exists(const.SUPPLDIRS[topo_dataset]):
            topodata_loc = const.SUPPLDIRS[topo_dataset]    
            print_log.info('Found default location for {} topodata at\n{}'
                      .format(topo_dataset, topodata_loc))
        
    try:
        access = geonum.TopoDataAccess(topo_dataset, local_path=topodata_loc)
        return access.get_data(lat0, lon0, lat1, lon1)
    except Exception as e:
        if try_etopo1 and not topo_dataset=='etopo1':
            print_log.warning('Failed to access topography data for {}. '
                           'Trying ETOPO1.\nError: {}'.format(topo_dataset, repr(e)))
            return get_topo_data(lat0, lon0, lat1, lon1, 
                                 topo_dataset='etopo1', 
                                 topodata_loc=topodata_loc,
                                 try_etopo1=False)
        raise
        
def get_topo_altitude(lat, lon, topo_dataset='srtm', topodata_loc=None,
                      try_etopo1=True):
    """Retrieve topographic altitude for a certain location
    
    Currently works only if :mod:`geonum` is installed. Supports topography
    datasets supported by geonum. These are currently (20 Feb. 19) srtm 
    (SRTM dataset, default, automatic access if online) and etopo1 
    (ETOPO1 dataset, lower resolution, must be available on local machine or
    server). 
    
    Parameters
    ----------
    lat : float
        latitude of coordinate
    lon : float
        longitude of coordinate
    topo_dataset : str
        name of topography dataset
    topodata_loc : str
        filepath or directory containing supported topographic datasets
    try_etopo1 : bool
        if True and if access fails via input arg `topo_dataset`, then try
        to access altitude using ETOPO1 dataset.
        
    Returns
    -------
    dict
        dictionary containing input latitude, longitude, altitude and 
        topographic dataset name used to retrieve the altitude.
    
    Raises
    ------
    ValueError
        if altitude data cannot be accessed
    """
    return get_topo_data(lat, lon, topo_dataset=topo_dataset, 
                         topodata_loc=topodata_loc,
                         try_etopo1=try_etopo1)(lat, lon)
    
def calc_distance(lat0, lon0, lat1, lon1, alt0=None, alt1=None,
                  auto_altitude_srtm=False):
    """Calculate distance between two coordinates
    
    Parameters
    ----------
    lat0 : float
        latitude of first point in decimal degrees
    lon0 : float
        longitude of first point in decimal degrees
    lat1 : float 
        latitude of secondpoint in decimal degrees
    lon1 : float
        longitude of second point in decimal degrees
    alt0 : :obj:`float`, optional
        altitude of first point in m
    alt1 : :obj:`float`, optional
        altitude of second point in m
    auto_altitude_srtm : bool
        if True, then all altitudes that are unspecified are set to the 
        corresponding topographic altitude of that coordinate, using SRTM 
        (requires geonum to be available and works only for coordinates where
        SRTM topographic data is accessible).
        
    Returns
    -------
    float
        distance between points in km
    """
    if not GEONUM_AVAILABLE and auto_altitude_srtm:
        raise ModuleNotFoundError('Require Geonum library for accessing '
                                  'topographic altitude using SRTM database')
    if GEONUM_AVAILABLE:
        import geonum
        p0 = geonum.GeoPoint(lat0, lon0, alt0, 
                             auto_topo_access=auto_altitude_srtm)
        p1 = geonum.GeoPoint(lat1, lon1, alt1, 
                             auto_topo_access=auto_altitude_srtm)
        if auto_altitude_srtm:
            if p0.altitude_err == p0._ALTERR_DEFAULT:
                raise ValueError('Failed to access topographic height for coord '
                                 '{} using SRTM topographic database'.format(p0))
            elif p1.altitude_err == p1._ALTERR_DEFAULT:
                raise ValueError('Failed to access topographic height for coord '
                                 '{} using SRTM topographic database'.format(p1))
        return (p0 - p1).magnitude
    else:
        logger.warning('geonum is not installed, computing approximate '
                       'distance using haversine formula')
        hordist = haversine(lat0, lon0, lat1, lon1)
        if alt0 == None:
            alt0 = 0
        if alt1 == None:
            alt1 = 0 
        return np.linalg.norm((hordist, (alt0 - alt1)/1000))
        
    
def is_within_radius_km(lat0, lon0, lat1, lon1, maxdist_km, alt0=0, alt1=0, 
                        **kwargs):
    """Checks if two lon/lat coordinates are within a certain distance to each other
    
    Parameters
    ----------
    lat0 : float
        latitude of first point in decimal degrees
    lon0 : float
        longitude of first point in decimal degrees
    lat1 : float 
        latitude of second point in decimal degrees
    lon1 : float
        longitude of second point in decimal degrees
    maxdist_km : float 
        maximum distance between two points in km
    alt0 : float
        altitude of first point in m
    alt1 : float
        altitude of second point in m
    
    Returns
    -------
    bool
        True, if coordinates are within specified distance to each other, else
        False
        
    """
    dist = calc_distance(lat0, lon0, lat1, lon1, alt0=alt0, alt1=alt1)
    if dist <= maxdist_km:
        return True
    return False

def haversine(lat0, lon0, lat1, lon1, earth_radius=6371.0):
    """Haversine formula
    
    Approximate horizontal distance between 2 points assuming a spherical 
    earth using haversine formula.
    
    Note
    ----
    This code was copied from geonum library (date 12/11/2018, J. Gliss)
    
    Parameters
    ----------
    lat0 : float
        latitude of first point in decimal degrees
    lon0 : float
        longitude of first point in decimal degrees
    lat1 : float 
        latitude of second point in decimal degrees
    lon1 : float
        longitude of second point in decimal degrees
    earth_radius : float 
        average earth radius in km, defaults to 6371.0
        
    Returns
    --------
    float
        horizontal distance between input coordinates in km 
    """
    hav = lambda d_theta: np.sin(d_theta / 2.0) ** 2
    
    d_lon = np.radians(lon1 - lon0)
    d_lat = np.radians(lat1 - lat0)
    lat0 = np.radians(lat0)
    lat1 = np.radians(lat1)
 
    a = hav(d_lat) + np.cos(lat0) * np.cos(lat1) * hav(d_lon)
    c = 2 * np.arcsin(np.sqrt(a))
 
    return earth_radius * c

if __name__ == '__main__':
    lat = 50.7
    lon = 8.2
    
    print(get_topo_altitude(0,0,try_etopo1=True))
    
    a0 = get_topo_altitude(lat, lon, 'srtm', try_etopo1=False)
    a1 = get_topo_altitude(lat, lon, 'etopo1', try_etopo1=False)
    
    print(a0, a1)
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Heatmap plotting functionality
"""
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
from seaborn import heatmap

def df_to_heatmap(df, cmap="bwr", center=None, low=0.3, high=0.3, vmin=None,
                  vmax=None, color_rowwise=True,
                  normalise_rows=False, normalise_rows_how='median',
                  normalise_rows_col=None,
                  annot=True, table_name="", num_digits=0, ax=None, 
                  figsize=(12,12), cbar=False, xticklabels = None, 
                  yticklabels = None, xlabel= '',  ylabel = '', circle = None, **kwargs):
    

    """Plot a pandas dataframe as heatmap
    
    Parameters
    ----------
    df : DataFrame
        table data
    cmap : str
        string specifying colormap to be used
    center : float
        value that is mapped to center colour of colormap (e.g. 0)
    low : float
        Extends lower range of the table values so that when mapped to  the 
        colormap, it’s entire range isn’t used. E.g. 0.3 roughly corresponds 
        to colormap crop of 30% at the lower end.
    high : float
        Extends upper range of the table values so that when mapped to the 
        colormap, it’s entire range isn’t used. E.g. 0.3 roughly corresponds 
        to colormap crop of 30% at the upper end.
    color_rowwise : bool
        if True, the color mapping is applied row by row, else, for the whole
        table
    normalise_rows : bool
        if True, the table is normalised in a rowwise manner either using the
        mean value in each row (if argument ``normalise_rows_col`` is 
        unspecified) or using the value in a specified column. 
    normalise_rows_how : str
        aggregation string for row normalisation. Choose from ``mean, median,
        sum``. Only relevant if input arg ``normalise_rows==True``.
    normalise_rows_col : int, optional
        if provided and if prev. arg. ``normalise_rows==True``, then the 
        corresponding table column is used for normalisation rather than 
        the mean value of each row
    
    annot : bool
        if True, the table values are printed into the heatmap
    table_name : str
        title of plot and label of colorbar (if colorbar is included, see cbar
        option)
    num_digits : int
        number of digits printed in heatmap annotation
    ax : axes
        matplotlib axes instance used for plotting, if None, an axes will be
        created
    figsize : tuple
        size of figure for plot
    cbar : bool
        if True, a colorbar is included
    xticklabels : List[str]
        List of string labels.
    yticklabels : List[str]
        List of string labels.
        
    Returns
    -------
    axes
        plot axes instance
    """
    
    # Old vesion only for 
    if isinstance(df.columns, pd.MultiIndex):
        # If pandas is a instance of multicolumns make sure that it only ha one column.
        if len(df.columns.levels) > 1:
             raise AttributeError("Heatmaps can only be created for "+
                                 "single column tabular data (e.g. Bias or "+
                                 "RMSE) with a partly unstacked MultiIndex or a "+
                                 "single index DataFrame. "+
                                 "Not MulitiIndex of {} ".format(len(df.columns.levels[0]))+
                                 " columns which you provided. "+
                                 "Please extract a column. ")  
            
    if circle:
        raise NotImplementedError('Adding circles to heatmap is not implemented yet.')

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    else:
        fig = ax.figure
    cbar_kws = {}
    
    if normalise_rows:
        if normalise_rows_col is not None:
            norm_ref = df.values[:, normalise_rows_col]
            table_name += " (norm. col. {})".format(normalise_rows_col)
        else:
            if normalise_rows_how == 'mean':
                norm_ref = df.mean(axis=1)
            elif normalise_rows_how == 'median':
                norm_ref = df.median(axis=1)
            elif normalise_rows_how == 'sum':
                norm_ref = df.sum(axis=1)
            else:
                raise ValueError('Invalid input for normalise_rows_how ({}). '
                                 'Choose from mean, median or sum'.format(normalise_rows_how))
            table_name += " (norm. row {})".format(normalise_rows_how)
        df = df.subtract(norm_ref, axis=0).div(norm_ref, axis=0)
        num_fmt = ".0%"
        
        cbar_kws['format'] = FuncFormatter(lambda x, pos: '{:.0%}'.format(x))
        #df = df.div(df.max(axis=1), axis=0)
    if color_rowwise:
        df_hm = df.div(abs(df).max(axis=1), axis=0)
    else:
        df_hm = df

    cbar_kws['label'] = table_name
    if annot is True:
        annot = df.values

    if vmin is None:
        vmin = df_hm.min().min() * (1-low)
    elif vmax is None:
        vmax = df_hm.max().max() * (1+high)
    
    # If the user needs too many displays 
    if num_digits < 5:
        num_fmt = ".{}f".format(num_digits)
    else:
        num_fmt = ".4g"

    ax = heatmap(df_hm, cmap=cmap, center = center, annot=annot, ax=ax, # changes this from df_hm to df because the annotation and colorbar didn't work.
                 fmt=num_fmt, cbar=cbar, cbar_kws=cbar_kws,
                 vmin=vmin, vmax=vmax, **kwargs)
        
    ax.set_title(table_name, fontsize=16)
    
    if yticklabels:
        ax.set_yticklabels(yticklabels, rotation=45, ha='left')
    
    if xticklabels:
        ax.set_xticklabels(xticklabels, rotation=45, ha='right')
    else:
        xticklabels = ax.get_xticklabels()
        ax.set_xticklabels(xticklabels, rotation=45, ha='right')
    
    if xlabel:
        ax.set_xlabel(xlabel, fontsize = 12)
    else:
        ax.set_xlabel('') 
    
    if ylabel:
        ax.set_ylabel(ylabel, fontsize = 12)
    else:
        ax.set_ylabel('') 
    
    fig.tight_layout()
    
    return ax
